# Copyright (C) 2016-2017 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snf_django.management.commands import SynnefoCommand, CommandError
from synnefo.db import transaction
from synnefo.management.common import get_resource
from synnefo.db.models import IPAddress


class Command(SynnefoCommand):
    help = """
Change owner of NICs and IPs to match the owner of the attached VM.
Also change project where applicable."""
    args = "<serverid> <old_ownerid>"

    @transaction.atomic
    def fix_nics_and_ips(self, serverid, old_owner):
        server = get_resource("server", serverid, for_update=True)
        new_owner = server.userid
        nics = server.nics.filter(userid=old_owner).select_for_update()
        if not nics:
            self.stdout.write("No NIC found.\n")
            return
        for nic in nics:
            nic.userid = new_owner
            nic.save()
            self.stdout.write(
                "Changed the owner of port '%s' from '%s' to '%s'.\n" %
                (nic.id, old_owner, new_owner))
        for ip in IPAddress.objects.filter(
                nic__machine=server, userid=old_owner).\
                select_for_update().select_related("nic"):
            old_project = ip.project
            ip.userid = new_owner
            if old_project is not None:
                ip.project = new_owner
            ip.save()
            self.stdout.write(
                "Changed the owner of IP '%s' from '%s' to '%s'.\n" %
                (ip, old_owner, new_owner))
            if old_project is not None:
                self.stdout.write(
                    "Changed the project of IP '%s' from '%s' to '%s'.\n" %
                    (ip, old_project, new_owner))

        self.stdout.write("WARNING: User quotas should be out of sync now,"
                          " run `snf-manage reconcile-resources-cyclades'"
                          " to review and update them.\n")

    def handle(self, *args, **options):
        if len(args) != 2:
            raise CommandError("Please provide <serverid> and <old_ownerid>")

        self.fix_nics_and_ips(args[0], args[1])
