# Copyright (C) 2016 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from optparse import make_option
from snf_django.management import utils
from snf_django.management.commands import SynnefoCommand, CommandError
from okeanos_policy.overquota.quota_policy import quota_policy, apply_prepared_actions
from okeanos_policy.overquota import config as cf


class Command(SynnefoCommand):
    help = """
    Get overquota memberships and perform actions according to the policy.
    """

    command_option_list = (
        make_option("--exclude-users",
                    default=[],
                    help=("Exclude list of users from overquota notification")
                    ),
        make_option("--projects",
                    help=("Search overquota users to the given list"
                          "of projects, e.g uuid1,uuid2")),
        make_option("--exclude-projects",
                    default=[],
                    help=("Exclude list of projects from overquota"
                          " notification")),
        make_option('--states',
                    help='Search memberships with the given overquota states'),
        make_option('--run',
                    action='store_true',
                    default=False,
                    help='Proceed the transition and action of memberships'
                    ' according to the policy'),
        make_option('--repeat',
                    action='store_true',
                    default=False,
                    help='Repeat action of current states of memberships'),
        make_option('--show-policy',
                    action='store_true',
                    default=False,
                    help='Show specified policy'),
        make_option('--uuids',
                    default=True,
                    choices=['True', 'False', 'true', 'false'],
                    help='Show uuids of projects and users [Default: True]',
                    metavar='True|False'),
        make_option('--names',
                    default=False,
                    choices=['True', 'False', 'true', 'false'],
                    help='Show names of projects and users [Default: False]',
                    metavar='True|False'),
        make_option('--show-actions',
                    default=False,
                    action='store_true',
                    help='Show next actions to be performed analytically'),
        make_option('--actions',
                    help=('Set actions to be performed; need to also specify '
                          'a single state'),
                    metavar='ACTION'),
        )

    def validate_states(self, states):
        if states is None:
            return cf.OVERQUOTA_STATES
        states = states.split(',')
        prefix_states = {state.split(cf.SEP)[0] for state in states}
        invalid_states = prefix_states.difference(cf.POLICY_CHAIN.keys())
        if invalid_states:
            msg = 'Unknown overquota state "%s"' % (invalid_states.pop())
            raise CommandError(msg)
        return states

    def validate_projects(self, projects):
        if projects is None:
            return set()
        return set(projects.split(','))

    def validate_uuids_names(self, uuids, names):
        if not uuids and not names:
            msg = 'You must specify columns to show'
            raise CommandError(msg)
        return uuids, names

    def set_state_actions(self, states, actions):
        if actions is None:
            return
        actions = actions.split(',')
        if len(states) != 1:
            raise CommandError("Need to specify exactly one state")
        state = states[0]
        category = state.split(cf.SEP)[0]
        cf.POLICY_CHAIN[category]['actions'] = actions

    def handle(self, *args, **options):
        states = self.validate_states(options['states'])
        self.set_state_actions(states, options['actions'])

        output_format = options['output_format']
        if options['show_policy']:
            policy = [(state, info['policy'] or 'immediate', info['actions'])
                      for state, info in cf.POLICY_CHAIN.iteritems()]
            utils.pprint_table(self.stdout, policy,
                               ['state type', 'scheduled', 'actions'],
                               output_format, title='Policy')
            exit()

        projects = self.validate_projects(options['projects'])
        exclude_projects = set(cf.EXCLUDE_PROJECTS)
        ex_projects = options['exclude_projects']
        if ex_projects:
            exclude_projects.update(ex_projects.split(','))

        exclude_users = set(cf.EXCLUDE_USERS)
        ex_users = options['exclude_users']
        if ex_users:
            exclude_users.update(ex_users.split(','))
        run, repeat = options['run'], options['repeat']
        listing = not (repeat or run)
        show_actions = options['show_actions']
        uuids, names = self.validate_uuids_names(
            utils.parse_bool(options['uuids']),
            utils.parse_bool(options['names']))
        (transitions, trans_labels), (next_actions, act_labels) = quota_policy(
            states, projects, exclude_projects, exclude_users, repeat, listing,
            run, True, uuids, names)
        utils.pprint_table(self.stdout, transitions, trans_labels,
                           output_format, title='Transitions')
        if show_actions and next_actions:
            self.stdout.write('\n')
            for i, next_action in enumerate(next_actions):
                utils.pprint_table(self.stdout, [next_action],
                                   act_labels[i],
                                   output_format, title='Action #%s' % (i + 1),
                                   vertical=True)
        if run and next_actions:
            self.stdout.write('Applying actions...\n')
            apply_prepared_actions()
