# Copyright (C) 2016-2017 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from optparse import make_option

from snf_django.management.commands import SynnefoCommand, CommandError
from synnefo.management.common import get_resource
from synnefo.db.models import Volume, IPAddress
from synnefo.logic import servers
from synnefo.logic import ips as ip_logic
from synnefo.volume import volumes as volume_logic
from snf_django.lib.api import Credentials

credentials = Credentials("snf-manage", is_admin=True)


class Command(SynnefoCommand):
    help = """
Reassign server to project. Can choose to propagate to volumes and IPs."""

    args = "<serverid> <projectid>"
    command_option_list = (
        make_option('--volumes',
                    action="store_true",
                    default=False,
                    help="Also change project of attached volumes"),
        make_option('--ips',
                    action="store_true",
                    default=False,
                    help="Also change project of attached IPs"),
    )

    def reassign_server(self, serverid, projectid):
        server = get_resource("server", serverid)
        self.stdout.write(
            "Reassigning server %s from project %s to %s.\n" %
            (server, server.project, projectid))
        servers.reassign(serverid, projectid, server.shared_to_project,
                         credentials=credentials)

    def reassign_volumes(self, serverid, projectid):
        server = get_resource("server", serverid)
        volumes = Volume.objects.filter(
            deleted=False, machine_id=serverid, userid=server.userid
        ).exclude(index=0)
        for volume in volumes:
            self.stdout.write(
                "Reassigning volume %s from project %s to %s.\n" %
                (volume, volume.project, projectid))
            volume_logic.reassign_volume(
                volume.id, projectid, volume.shared_to_project,
                credentials=credentials)

    def reassign_ips(self, serverid, projectid):
        server = get_resource("server", serverid)
        ips = IPAddress.objects.filter(
            nic__machine=server, userid=server.userid, deleted=False,
            floating_ip=True)
        for ip in ips:
            self.stdout.write(
                "Reassigning floating IP %s from project %s to %s.\n" %
                (ip, ip.project, projectid))
            ip_logic.reassign_floating_ip(
                ip.id, projectid, ip.shared_to_project,
                credentials=credentials)

    def handle(self, *args, **options):
        volumes_flag = options["volumes"]
        ips_flag = options["ips"]

        if len(args) != 2:
            raise CommandError("Please provide <serverid> and <projectid>")

        serverid = args[0]
        projectid = args[1]
        self.reassign_server(serverid, projectid)
        if volumes_flag:
            self.reassign_volumes(serverid, projectid)
        if ips_flag:
            self.reassign_ips(serverid, projectid)
