# Copyright (C) 2017 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from optparse import make_option
from snf_django.management.commands import SynnefoCommand
from snf_django.management import utils
from okeanos_policy.shibboleth import enforce, config
from astakos.im import transaction
from logging.config import dictConfig

from synnefo.settings import SNF_MANAGE_LOGGING_SETUP
SNF_MANAGE_LOGGING_SETUP['handlers']['shib_file'] = {
    'class': 'logging.FileHandler',
    'filename': '/var/log/Expired_Users.log',
    'formatter': 'simple'
    }
SNF_MANAGE_LOGGING_SETUP['handlers']['console']['level'] = 'WARNING'
SNF_MANAGE_LOGGING_SETUP['handlers']['term'] = {
    'class': 'logging.StreamHandler',
    'formatter': 'simple',
    'stream': 'ext://sys.stdout'}

SNF_MANAGE_LOGGING_SETUP['loggers']['shibboleth'] = {
    'handlers': ['shib_file', 'term'],
    'level': 'INFO',
    }
dictConfig(SNF_MANAGE_LOGGING_SETUP)

logger = logging.getLogger('shibboleth')


class Command(SynnefoCommand):
    help = """
Notify users that have not logged in using shibboleth for a specified period
and suspend them according to the admin-defined policy

"""

    command_option_list = (
        make_option("--run",
                    help='Execute the actual operations',
                    default=False,
                    action='store_true'),
        make_option("--recover",
                    help=('Recover users that have re-logged '
                          'in using shibboleth [Default: True]'),
                    default='True',
                    choices=['True', 'False'],
                    metavar='True|False'),
        make_option("--expired",
                    help=('Handle users with expired shibboleth account '
                          '[Default: True]'),
                    default='True',
                    choices=['True', 'False'],
                    metavar='True|False'),
        make_option("--migrate-ok-policy-enforce",
                    dest="migrate",
                    help=("Migrate from old script 'ok_policy_enforce.py'. "
                          "It assumes recover=True, expired=False"),
                    default=False,
                    action='store_true'),
        make_option("--show-policy",
                    help="Show the specified policy",
                    default=False,
                    action='store_true'),
    )

    @transaction.atomic
    def handle_recovering(self):
        logger.info("Handling recovering users...")
        loggedin = enforce.get_loggedin_users()
        for user in loggedin:
            enforce.recover_loggedin_user(user, self.dry_run)
        return loggedin

    @transaction.atomic
    def handle_expired(self):
        logger.info("Handling expired users...")
        expired = enforce.get_expired_users()
        for user in expired:
            auth = user.get_auth_provider('shibboleth')
            auth = auth._instance
            user = auth.user
            logger.info("User %s (%s), last logged in using shibboleth on %s",
                        user.email, user.uuid, auth.last_login_at)
            shib_values = enforce.initialize_expired_auth(auth, self.dry_run)
            enforce.check_user(auth.user, self.dry_run, shib_values)

    @transaction.atomic
    def handle_migration(self, exclude=None):
        logger.info("Handling states from 'ok_policy_enforce.py'...")
        users = enforce.get_users_to_migrate(exclude=exclude)
        for user in users:
            enforce.migrate_user(user, self.dry_run)

    def print_policy(self):
        policy = config.POLICY_CHAIN
        l = []
        labels = ("state type", "scheduled in days", "actions")
        for name, data in policy.iteritems():
            if data.get('deprecated'):
                continue
            days = data['policy']
            if not days:
                days = 'immediate'
            l.append((name, days, data['actions']))
        utils.pprint_table(self.stdout, l, labels,
                           output_format=self.output_format,
                           title="Policy")
        self.stdout.write(
            "\nShibboleth is considered expired after %s "
            "days without login.\n" %
            config.SHIBBOLETH_EXPIRES_DAYS)

    def handle(self, *args, **options):
        self.output_format = options["output_format"]
        self.dry_run = not options["run"]
        if options["show_policy"]:
            self.print_policy()
            return

        MODE = "on" if self.dry_run else "off"
        logger.info("#### Dry run mode is %s" % MODE)

        if options["migrate"]:
            loggedin = self.handle_recovering()
            self.handle_migration(exclude=loggedin)
            return

        if utils.parse_bool(options["recover"]):
            self.handle_recovering()

        if utils.parse_bool(options["expired"]):
            self.handle_expired()
