# Copyright (C) 2016 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
from optparse import make_option

from astakos.im.models import Project
from astakos.im.notifications import build_notification
from astakos.im import settings
from snf_django.management.commands import SynnefoCommand

DEFAULT_WARN_DAYS = 15

SENDER = settings.SERVER_EMAIL
SUBJECT = "Warning: Project %s will be terminated soon"
TEMPLATE = "projects/project_termination_warning.txt"


def sendemail(recipient, subj, template, tvars):
    notification = build_notification(
        SENDER,
        [recipient],
        subj,
        template=template,
        dictionary=tvars)
    notification.send()


class Command(SynnefoCommand):
    help = "Warn project owners before project expiration"

    command_option_list = (
        make_option("--run",
                    help='Execute the actual operations',
                    default=False,
                    action='store_true'),
        make_option("--days",
                    help=('Warn owner DAYS before project expiration'
                          ' (default: %s)' % DEFAULT_WARN_DAYS),
                    default=DEFAULT_WARN_DAYS),
        )

    def handle(self, *args, **options):
        warn_days = int(options.get('days'))
        RUN = options.get('run')
        if not RUN:
            self.stdout.write("Running in dry run mode.\n")
        now = datetime.datetime.now()
        delta = datetime.timedelta
        from_date = now + delta(days=warn_days-1)
        to_date = now + delta(days=warn_days)
        projects = Project.objects.select_related('owner').filter(
            state=Project.NORMAL,
            end_date__gte=from_date,
            end_date__lt=to_date)
        for project in projects:
            owner = project.owner
            if not owner.is_active:
                self.stdout.write(
                    "Owner %s of %s is not active; skipping.\n" %
                    (owner, project))
                continue
            self.stdout.write(
                "Informing owner %s of %s (expiry date: %s).\n" %
                (owner, project, project.end_date))
            if RUN:
                subject = SUBJECT % project.realname
                sendemail(owner.email, subject, TEMPLATE, {
                        "project": project,
                        "user": owner,
                        "warn_days": warn_days,
                        "baseurl": settings.BASE_URL,
                        })
