# Copyright (C) 2016 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from collections import defaultdict
from astakos.im.models import ProjectMembership
from okeanos_policy.overquota import config as cf
from okeanos_policy.overquota import utils


OK_PROJECT_ACTIONS = ['REINSTATE', 'UNSUSPEND', 'MODIFY']
OK_MEMBERSHIP_ACTIONS = ['ENROLL', 'JOIN', 'ACCEPT', 'UNSUSPEND']
CHECK_PROJECT_ACTIONS = ['TERMINATE', 'SUSPEND', 'MODIFY']
CHECK_MEMBERSHIP_ACTIONS = ['REMOVE', 'SUSPEND', 'LEAVE']


def check_state_projects(projects, action):
    """
    Check the overquota state of the specified list of projects and decides
    based on the project action if the state of (overquota) users should be
    updated to `OK` or to 'OVERQUOTA'.

    For example, if given projects have been suspended then, the new state of
    the memberships of their overquota users is `OVERQUOTA`.

    Args:
        projects: List of projects to the get their overquota users.
        action: Action which is performed to the specified list of projects,
        e.g. TERMINATE, UNSUSPEND, etc.
    """
    memberships = ProjectMembership.objects.select_related(
        'project', 'person').filter(project__in=projects)
    overquota_users = utils.get_overquota_users(memberships, cf.SERVICES)
    for project in projects:
        user_ids = overquota_users.get(project.uuid, [])
        if user_ids and action in CHECK_PROJECT_ACTIONS:
            utils.update_overquota_state(user_ids, project.uuid,
                                         cf.OVERQUOTA_OVERQUOTA)
        if action in OK_PROJECT_ACTIONS:
            utils.update_overquota_state(
                user_ids, project.uuid,
                cf.OVERQUOTA_OK, exclude=True)


def check_state_membership(membership, action):
    """
    Check if given membership is overquota and based on the action which is
    performed on it, decide of if the state of (overquota) user should be
    updated to `OK` or to `OVERQUOTA`.

    For example, if membership is overquota and the action which was performed
    was LEAVE, the new state of membership is `OVERQUOTA`.

    Args:
        membership: Membership to check its overquota state.
        action: Action which is performed to the given state, e.g. LEAVE,
        SUSPEND.
    """
    is_overquota = utils.is_membership_overquota(membership, cf.SERVICES)
    state = None
    if action in OK_MEMBERSHIP_ACTIONS and not is_overquota:
        state = cf.OVERQUOTA_OK
    if action in CHECK_MEMBERSHIP_ACTIONS and is_overquota:
        state = cf.OVERQUOTA_OVERQUOTA
    if state:
        utils.update_overquota_state([membership.person.uuid],
                                     membership.project.uuid, state)
