# -*- coding: utf-8 -*-
# Copyright (C) 2017 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import subprocess
from collections import defaultdict
from astakos.im import settings
from astakos.im.management.commands._common import show_resource_value
from astakos.im.notifications import build_notification, NotificationError
from okeanos_policy.overquota import config as cf
from okeanos_policy.overquota import utils


LOGGER = logging.getLogger(__name__)


OK = cf.OVERQUOTA_OK
OVERQUOTA = cf.OVERQUOTA_OVERQUOTA


READABLE_RESOURCES = {
    'cyclades.ram': 'RAM',
    'cyclades.disk': 'Disks',
    'cyclades.cpu': 'CPUs',
    'cyclades.vm': 'VMs',
    'cyclades.floating_ip': 'IP addresses',
}

TRANSLATED_ACTIONS = {
    'DESTROY': 'Καταστρoφή',
    'SHUTDOWN': 'Απενεργοποίηση',
    'DETACH': 'Αποσύνδεση',
}

ACTIONS_PP = {
    'DESTROY': 'DESTROYED',
    'SHUTDOWN': 'SHUTDOWN',
    'DETACH': 'DETACHED',
}


TRANSLATED_REASONS = {
    'Modified': 'Τροποποιηθεί',
    'Suspended': 'Ανασταλεί',
    'Terminated': 'Τερματιστεί',
    'Removed': 'Αφαιρεθεί',
}


OVERQUOTA_NOTIF = {
    'subject':  'Attention: Exceeded quota',
    'template': 'overquota/project_membership_overquota_notification.txt',
}


OVERQUOTA_NOTIF_ADMIN = {
    'subject':  'Quota enforcement',
    'template': 'overquota/project_membership_overquota_notif_admin.txt',
}


OVERQUOTA_LOG_ADMIN = {
    'subject':  'Quota enforcement log',
    'template': 'overquota/project_membership_overquota_log_admin.txt',
}

USER_EMAIL_QUEUE = defaultdict(list)
ADMIN_EMAIL_VIOLATIONS_DATA = defaultdict(list)
ADMIN_ENFORCEMENT_QUEUE = defaultdict(list)

SENDER = settings.SERVER_EMAIL


def prepare_user_email(membership, next_state, **kwargs):
    """
    This function composes and sends an informative email to user who is
    overquota.

    This email contains a detailed report regarding with the resources, current
    usage and limit, as well as a list of the imminent actions to the user's
    resources.
    """
    severity = next_state.split('_')[0]
    usage = reformat_resource_usage(kwargs['usage'])
    resource_actions = kwargs['resource_actions']
    user = membership.person
    project = membership.project
    vm_destruction = ('DESTROY', 'cyclades.vm') in zip(*resource_actions)[1][1]
    resource_actions = [(days, actions)
                        for days, actions in resource_actions
                        if actions]

    notification = build_notification(
        SENDER,
        [user.email],
        OVERQUOTA_NOTIF['subject'],
        template=OVERQUOTA_NOTIF['template'],
        dictionary={'user': user, 'project': project, 'usage': usage,
                    'resource_actions': resource_actions,
                    'reason': utils.get_reason(membership),
                    'readable_resources': READABLE_RESOURCES,
                    'vm_destruction': vm_destruction,
                    'actions_pp': ACTIONS_PP,
                    'translated_actions': TRANSLATED_ACTIONS,
                    'translated_reasons': TRANSLATED_REASONS})
    USER_EMAIL_QUEUE[severity].append(notification)


def send_user_emails():
    for severity, notifications in USER_EMAIL_QUEUE.iteritems():
        for notification in notifications:
            try:
                notification.send()
            except NotificationError as e:
                LOGGER.error(e.message)


def prepare_admin_email(membership, next_state, **kwargs):
    """
    This function prepares an informative email to administrators.

    This email describes the actions that should be taken for the resources of
    user e.g. vm destruction.
    """
    usage = kwargs['usage']
    user = membership.person
    project = membership.project
    severity = next_state.split('_')[0]
    snf_command = create_snf_command(user.uuid, project.uuid, next_state,
                                     usage.keys())
    violation = {'user': user, 'project': project, 'usage': usage,
                 'command': snf_command,
                 'reason': utils.get_reason(membership)}
    ADMIN_EMAIL_VIOLATIONS_DATA[severity].append(violation)


def send_batch_admin_email():
    """
    This function sends an informative email to administrators.

    This email describes the actions that should be taken for the resources of
    user e.g. vm destruction.
    """
    for severity, data in ADMIN_EMAIL_VIOLATIONS_DATA.iteritems():
        if data:
            subject = "%s (%s)" % (
                OVERQUOTA_NOTIF_ADMIN['subject'], severity.lower())
            template = OVERQUOTA_NOTIF_ADMIN['template']
            dictionary = {'violations': data}
            send_admin_email(subject, template, dictionary)


def send_admin_email(subject, template, dictionary):
    _, admins = zip(*settings.ADMINS)
    try:
        notification = build_notification(
            SENDER,
            admins,
            subject,
            template=template,
            dictionary=dictionary)
        notification.send()
    except NotificationError as e:
        LOGGER.error(e.message)


def send_enforce_log(severity, data):
    subject = "%s (%s)" % (OVERQUOTA_LOG_ADMIN['subject'], severity.lower())
    template = OVERQUOTA_LOG_ADMIN['template']
    dictionary = {'logs': data}
    send_admin_email(subject, template, dictionary)


def create_snf_command(user_uuid, project_uuid, next_state,
                       overquota_resources):
    """
    This functions creates the synnefo command that will be executed by
    admins for the resources enforcement.

    Args:
        user_uuid: UUID of user who is overquota.
        project_uuid: UUID of project which is related with the overquota
        resources.
        next_state: Next overquota state of user.
        overquota_resources: List of overquota resources.
    Returns:
        String with the synnefo command to be executed.
    """
    soft_resources, resources = get_soft_hard_resources(
        next_state, overquota_resources)
    if not soft_resources and not resources:
        return None
    command = 'snf-manage enforce-resources-cyclades'
    user_opt = '--users=' + user_uuid
    project_opt = '--projects=' + project_uuid
    command += ' ' + user_opt + ' ' + project_opt
    if soft_resources:
        command += ' --soft-resources=' + ','.join(soft_resources)
    if resources:
        command += ' --resources=' + ','.join(resources)
    command += ' --fix --force'
    return command


def prepare_quota_enforce(membership, next_state, **kwargs):
    severity = next_state.split('_')[0]
    overquota_resources = kwargs['usage'].keys()
    snf_command = create_snf_command(
        membership.person.uuid, membership.project.uuid, next_state,
        overquota_resources)
    ADMIN_ENFORCEMENT_QUEUE[severity].append((snf_command, overquota_resources))


def run_snf_commands(snf_commands):
    N_DONE = 0
    N_FAILED = 0
    log = []
    for snf_command, overquota_resources in snf_commands:
        if snf_command is None:
            LOGGER.warning(
                "Resources '%s' cannot be enforced by snf-manage command"
                " tool at the moment" % (','.join(overquota_resources)))
            continue
        try:
            outp = subprocess.check_output(
                snf_command.split(' '), stderr=subprocess.STDOUT)
            log.append("%s\n%s" % (snf_command, outp))
            LOGGER.info('Enforcement command: %s', snf_command)
            N_DONE += outp.count("DONE")
            N_FAILED += outp.count("FAILED")
        except subprocess.CalledProcessError as e:
            LOGGER.error(
                "Enforcement command '%s' failed with code: %s, output: '%s'",
                snf_command, e.returncode, e.output)
    summary = "SUMMARY: %s actions done, %s actions failed.\n" % (
        N_DONE, N_FAILED)
    return [summary] + log


def apply_quota_enforce():
    for severity, snf_commands in ADMIN_ENFORCEMENT_QUEUE.iteritems():
        log = run_snf_commands(snf_commands)
        if log:
            send_enforce_log(severity, log)


def reformat_resource_usage(usage):
    for resource, info in usage.iteritems():
        usage[resource]['usage'] = show_resource_value(
            info['usage'], resource, 'auto')
        usage[resource]['limit'] = show_resource_value(
            info['limit'], resource, 'auto')
    return usage


def get_soft_hard_resources(next_state, resources):
    """
    Gets the resources that should be enforced softly, and the resources that
    should be enforced as usual.

    For more information, see the documentation of
    `enforce-resources-cyclades`.

    Args:
        next_state: Next state transition.
        resources: List of overquota resources.

    Returns:
        1) List of resources that should be handled softly.
        2) List of resources that should be handled as usual.
    """
    severity, _ = tuple(next_state.split('_'))
    if severity == cf.SOFT:
        resources = set(resources) - set(cf.ONLY_HARD_RESOURCES)
        soft_resources = resources & set(cf.HARD_RESOURCES)
        resources = set(resources) - soft_resources
    else:
        soft_resources = []
    return soft_resources, resources
