# Copyright (C) 2016 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict
from datetime import datetime
from astakos.im import transaction
from okeanos_policy.overquota import config as cf
from okeanos_policy.overquota import policy_actions as pa
from okeanos_policy.overquota import utils


OK = cf.OVERQUOTA_OK
OVERQUOTA = cf.OVERQUOTA_OVERQUOTA


@transaction.atomic
def quota_policy(states, project_ids, ex_project_ids, ex_user_ids,
                 repeat=False, listing=True, run=False, verbose=False,
                 uuids=True, names=False):
    """
    This functions retrieves overquota memberships according to the
    parameters and forms the transition of state for every membership according
    to the policy.

    Args:
        states: List of overquota states of retrieved memberships.
        project_ids: List of project uuids.
        ex_project_ids: List of project uuids to be excluded.
        ex_user_ids: List of user uuids to be excluded.
        repeat: True if action which is specified by policy should be repeated;
        False otherwise.
        listing: True if next transitions of memberships will be presented;
        False otherwise.
        run: True if actions will be performed according to the policy.
        verbose: True if a description of next actions should be included;
        False otherwise.
        uuids: True if uuids of projects and users should be presented;
        False otherwise.
        names: True if names of projects and users should be presented;
        False otherwise.
    """
    memberships = utils.get_memberships(states, project_ids, ex_project_ids,
                                        ex_user_ids)
    overquota_users = utils.get_overquota_users(
        memberships, cf.SERVICES, verbose=True, resource=cf.RESOURCES)
    transitions, actions_to_perform = check_memberships(
        memberships, overquota_users, repeat,
        apply_policy=run, verbose=verbose)
    return (utils.format_transitions(transitions, not (listing or repeat), run,
                                     uuids, names),
            utils.format_actions(actions_to_perform, uuids, names))


def check_memberships(memberships, overquota_users, repeat, apply_policy=True,
                      verbose=False):
    """
    This method gets a list of memberships and and applies the policy based
    on their overquota state and the overquota users per project.

    Args:
        memberships: List of memberships.
        overquota_users: Dictionary keyed by project that contains a dictionary
        keyed by user and its usage per resource.
        repeat: True if action which is specified by policy should be repeated;
        False otherwise.
        apply_policy: True if action must be performed according to the policy;
        False otherwise.
        verbose: True if a description of next actions should be included;
        False otherwise.
    Returns:
        A dictionary keyed by updated overquota state that contains information
        about the state's transitions for each user and project according to
        the policy.
    """
    transitions = []
    updated_memberships = defaultdict(lambda: defaultdict(lambda: []))
    actions_to_perform = []
    for membership in memberships:
        user_tuple = (membership.person.uuid, membership.person.email)
        project_tuple = (membership.project.uuid, membership.project.name)
        users_usage = overquota_users.get(membership.project.uuid, {})
        if membership.person.uuid in users_usage:
            policy, next_action = _construct_policy(
                membership, users_usage[membership.person.uuid], apply_policy,
                repeat, verbose)
            if next_action is not None:
                actions_to_perform.append(next_action)
            if policy.status == QuotaPolicy.DONE:
                updated_memberships[policy.next_overquota_state][
                    project_tuple[0]].append(user_tuple[0])
            transitions.append((
                project_tuple, user_tuple, membership.overquota_state,
                policy.next_overquota_state, policy.time_next_state,
                policy.status))
        elif membership.overquota_state != OK:
            updated_memberships[OK][project_tuple[0]].append(user_tuple[0])
            transitions.append((
                project_tuple, user_tuple, membership.overquota_state,
                OK, 0, QuotaPolicy.DONE if apply_policy else None))
    if apply_policy:
        utils.update_membership_state(updated_memberships)
    return transitions, actions_to_perform


def _construct_policy(membership, usage, apply_policy, repeat, verbose):
    """
    Construct policy and get executable actions of given membership.

    Args:
        memberships: Membership which is overquota to constuct its policy.
        usage: Current usage of user specified by membership.
        repeat: True if action which is specified by policy should be repeated;
        False otherwise.
        apply_policy: True if action must be performed according to the policy;
        False otherwise.
        verbose: True if a description of next actions should be included;
        False otherwise.

    Returns:
        1) Constructed policy of membership.
        2) A list of tuples which descibes the next executable actions
        according to policy.
    """
    policy = QuotaPolicy(
        membership.overquota_state, usage.keys(),
        membership.overquota_date, repeat)

    next_actions = policy.get_next_actions(membership) if verbose else None
    if apply_policy:
        policy.prepare_policy_application(membership, usage)
    return policy, next_actions


POLICY_PREPARE_ACTIONS = {
    'inform_user': pa.prepare_user_email,
    'inform_admin': pa.prepare_admin_email,
    'enforce': pa.prepare_quota_enforce,
}

POLICY_APPLY_ACTIONS = {
    'inform_user': pa.send_user_emails,
    'inform_admin': pa.send_batch_admin_email,
    'enforce': pa.apply_quota_enforce,
}


def apply_prepared_actions():
    for policy, action in POLICY_APPLY_ACTIONS.iteritems():
        if action is None:
            continue
        action()


def split_overquota_state(overquota_state):
    """
    This function splits a specified overquota state by its category and the
    number of transitions.

    For example, SOFT_USER:1 is splitted by SOFT_USER and 1.

    If given overquota state is `OVERQUOTA` then number of transitions is not
    defined by definition.

    Args:
        overquota_state: Overquota state of a memberships which is going to
        be splitted.
    Returns:
        1) Category of overquota state.
        2) Number of transitions for the current state category.
    """
    if overquota_state in [OK, OVERQUOTA]:
        return overquota_state, 0
    state, notice = tuple(overquota_state.split(cf.SEP))
    return state, int(notice)


class QuotaPolicy(object):
    FAILED = 'failed'
    DONE = 'done'

    """
    This class defines the policy which is applied and designed for the
    quota reclamation.

    It is responsible for the handling of membership's overquota state and
    decides if an action should be perfomed regarding with overquota resources
    of a user as well as a transition to the next state.

    Actions and dates when a transition should be perfomed, are defined by
    settings which form current policy.
    """
    def __init__(self, overquota_state, overquota_resources, last_update,
                 repeat=False):
        self.current_category, self.transitions = split_overquota_state(
            overquota_state)
        self.overquota_resources = overquota_resources
        self.last_update = last_update
        self.repeat = repeat
        self.next_category = self.get_next_state_category()
        self.set_num_of_transitions()
        self.next_overquota_state = self.next_category + (
            cf.SEP + str(self.transitions) if self.transitions else '')
        self.time_next_state = self.get_time_next_state()
        self.status = None
        self.soft_time_period, self.hard_time_period = self.get_time_period()

    def get_next_state_category(self):
        """
        Get the next state based on the current state, notice and overquota
        resources.

        Typically, when the life cycle of a state ends, then go to the next
        state except than the case where overquota resource list contains only
        resources that only soft actions can be enforced and we cannot go the
        next state of hard warns.

        Returns:
            Next state of overquota user.
        """
        if self.repeat:
            return self.current_category
        policy = cf.POLICY_CHAIN[self.current_category]['policy']
        if self.transitions >= len(policy) and not\
                (self.current_category == cf.AMBIGUOUS_TRANSITION and (
                        not set(cf.HARD_RESOURCES).intersection(
                            self.overquota_resources))):
            # If resource list contains only resources that only soft action
            # can be applied then don't go to the next state.
            states = cf.POLICY_CHAIN.keys()
            if self.current_category == states[-1]:
                return self.current_category
            next_category = states[states.index(self.current_category) + 1]
            return next_category or self.current_category
        return self.current_category

    def set_num_of_transitions(self):
        """
        Set the number of transitions for the current state category.

        Typically, a number of transitions can be perfomed in the same category
        of a state according to the policy.

        For example, SOFT_USER can support 3 transitions, SOFT_USER:1,
        SOFT_USER:2, etc.

        Each transition defines an action which was perfomed for the overquota
        user, e.g send email to admin. Therefore, SOFT_USER:3 means that an
        action was perfomed three times and the user is on the 'SOFT_USER'
        category. Concatenation of number of actions and category of state is
        overquota state of user.
        """
        policy = cf.POLICY_CHAIN[self.next_category]['policy']
        if self.repeat:
            self.transitions = self.transitions\
                    if self.transitions <= len(policy) else len(policy)
            return
        if self.current_category == self.next_category:
            self.transitions = len(policy)\
                    if self.transitions + 1 > len(policy)\
                    else self.transitions + 1
        else:
            self.transitions = None if len(policy) == 0 else 1

    def get_time_next_state(self):
        """
        Get the number of days of the next state transition.

        Returns:
            Number of days of the next transition.
        """
        next_policy = cf.POLICY_CHAIN[self.next_category]['policy']
        if len(next_policy) == 0:
            return 0
        diff_days = (datetime.now() - self.last_update).days
        next_notice_day = next_policy[self.transitions - 1]
        return max(next_notice_day - diff_days, 0)

    def get_time_period(self):
        """
        Get time period for the next actions of ADMIN regarding to the
        overquota resources of a user.

        Returns:
            Time periods for the next actions of ADMIN, e.g. SHUTDOWN VM's.
        """
        hard = bool(set(cf.HARD_RESOURCES).intersection(
            self.overquota_resources))
        soft_exp_day = cf.SOFT_ADMIN_DAYS[-1]
        if not hard:
            return (self._calculate_time_period(soft_exp_day, True), None)
        hard_exp_day = cf.HARD_ADMIN_DAYS[-1]
        return (self._calculate_time_period(soft_exp_day, True),
                self._calculate_time_period(hard_exp_day, False))

    def _calculate_time_period(self, exp_day, soft):
        """
        Get the time period  for the next action of
        admin e.g. shutdown of VM's.

        If current state of user is 'HARD' then it is considered that `SOFT`
        action has been performed by ADMIN.

        Args:
            exp_day: Day when admin performs action.
            soft: True if action is considered `SOFT`, e.g. detach floating
            ip's' False otherwise.

        Returns:
            Time period for the next action of ADMIN.
        """
        if self.last_update is None or (
                cf.HARD in self.next_category and soft):
            return None
        if self.next_category == OVERQUOTA:
            return exp_day
        return max(exp_day - (datetime.now() - self.last_update).days, 0)

    def define_resource_actions(self):
        """
        Get sets of resources and the actions which are going to be performed
        according to the policy.

        For example, SOFT action for the VM's is the shutdowning of them,
        whereas the HARD action is the destruction of them.

        Returns:
            1) A set of tuples with the SOFT actions for each overquota
            resource.
            2) A set of tuples with the HARD actions for each overquota
            resource.
        """
        soft_actions = set()
        hard_actions = set()
        resource_actions = {
            cf.SOFT: soft_actions,
            cf.HARD: hard_actions
        }
        severity = self.next_category.split('_')[0]
        for resource in self.overquota_resources:
            try:
                resource_actions[severity].add(
                    cf.RESOURCE_ACTIONS[(severity, resource)])
            except KeyError:
                pass
            if severity == cf.SOFT:
                try:
                    hard_actions.add(
                        cf.RESOURCE_ACTIONS[(cf.HARD, resource)])
                except KeyError:
                    pass
        return soft_actions, hard_actions

    def should_run_policy(self):
        """
        Check if action specified by policy should be performed.

        Returns:
            True if policy should be run; False otherwise.
        """
        return not (self.time_next_state > 0 and not self.repeat)

    def get_next_actions(self, membership):
        """
        This method gets a list of actions which can be executed right now
        according to the policy.

        Args:
            membership: Overquota membership.
        Returns:
            List of tuples which contains data that describe action to be
            exetuted.
        """
        if not self.should_run_policy():
            return
        actions = cf.POLICY_CHAIN[self.next_category]['actions']
        project = membership.project
        user = membership.person
        pr_tuple = project.uuid, project.name
        usr_tuple = user.uuid, user.email
        next_actions = []
        for action in actions:
            next_action = (pr_tuple, usr_tuple, action)
            if action != 'inform_user':
                next_action += (pa.create_snf_command(
                    membership.person.uuid, membership.project.uuid,
                    self.next_category, self.overquota_resources),)
            else:
                soft_actions, hard_actions = self.define_resource_actions()
                resource_actions = [(self.soft_time_period, soft_actions),
                                    (self.hard_time_period, hard_actions)]
                next_action += utils.format_resource_actions(resource_actions)
            next_actions.append(next_action)
        return next_actions

    def prepare_policy_application(self, membership, usage):
        """
        This method applies policy according to the specified settings, e.g
        send email to admin, enforce quotas etc.

        Args:
            membership: A membership which is overquota.
            usage: Dictionary keyed by resource which contains the actual usage
            and limit.
        """
        if not self.should_run_policy():
            return

        self.status = QuotaPolicy.DONE
        actions = cf.POLICY_CHAIN[self.next_category].get('actions')
        soft_actions, hard_actions = self.define_resource_actions()
        resource_actions = [(self.soft_time_period, soft_actions),
                            (self.hard_time_period, hard_actions)]
        for action in actions:
            POLICY_PREPARE_ACTIONS[action](
                membership, self.next_category,
                usage=usage,
                resource_actions=resource_actions)
