# Copyright (C) 2016 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
import operator
from collections import defaultdict
from datetime import datetime
from django.db.models import Q, F
from astakos.im.models import Project, ProjectMembership
from astakos.im import quotas as qu
from okeanos_policy.overquota import config as cf


ACTION_LABELS = ['action_type']


LABELS = ['previous_state', 'new_state', 'status']


LISTING_LABELS = ['current_state', 'next_state', 'scheduled']


def format_transitions(transitions, only_policy, run, uuids, names):
    """
    Format given dictionary to a list of tuples.

    Args:
        transitions: List of tuples that describe transitions.
        only_policy: True if executable transitions are taken account of;
        False otherwise.
        run: True if transition was performed; False otherwise.
        uuids: True if uuids of projects and users should be presented;
        False otherwise.
        names: True if names of projects and users should be presented;
        False otherwise.
    Returns:
        1) A list of tuples representation of the given dictionary.
        2) A list of corresponding labels for each field.
    """
    assert uuids or names
    formatted_transitions = []
    for project, user, current_state, next_state, time, status in transitions:
        if only_policy and time != 0:
            continue
        data = (current_state, next_state)
        data += (status,) if run else (str(time) + ' days',)
        pr_uuid, pr_name = project
        user_uuid, user_name = user
        if names:
            data = (pr_name, user_name) + data
        if uuids:
            data = (pr_uuid, user_uuid) + data
        formatted_transitions.append(data)
    return formatted_transitions, get_labels(LABELS if run else LISTING_LABELS,
                                             uuids, names)


def format_actions(actions_to_perform, uuids, names):
    """
    Format next actions to be performed to a list of tuples.

    Args:
        actions_to_perform: A list of tuples with the next actions to be
        performed.
        uuids: True if uuids of projects and users should be presented;
        False otherwise.
        names: True if names of projects and users should be presented;
        False otherwise.
    Returns:
        1) A list of tuple representation of next actions.
        2) A list of lists with the corresponding labels for each field.
    """
    formatted_actions = []
    labels = []
    for action in itertools.chain(*actions_to_perform):
        assert 4 <= len(action) <= 5
        data = action[2:]
        project, usr = action[0:2]
        if uuids:
            data = (project[0], usr[0]) + data
        if names:
            data = (project[1], usr[1]) + data
        formatted_actions.append(data)
        labels.append(get_labels(ACTION_LABELS, uuids, names, action[2]))
    return formatted_actions, labels


def format_resource_actions(resource_actions):
    """
    Format resource actions to readable tuple of strings.

    Args:
        :resource_actions: List of tuples that contains actions on resources.
    Returns:
        Tuple of string representation of actions.
    """
    formatted_data = []
    for time, actions in resource_actions:
        if time is not None:
            actions = [" ".join(a) for a in actions]
            actions = ", ".join(actions)
            formatted_data.append('in ' + (str(time) + ' days ' + actions))
        else:
            formatted_data.append('-')
    return tuple(formatted_data)


def get_labels(labels, uuids, names, action=None):
    if names:
        labels = ['project_name', 'user_name'] + labels
    if uuids:
        labels = ['project_uuid', 'user_uuid'] + labels
    if action is not None:
        labels += ['command'] if action != 'inform_user' else\
                ['soft_actions', 'hard_actions']
    return labels


def get_memberships(states, project_ids, ex_project_ids, ex_user_ids):
    """
    This functions get memberships which are associated with the specified
    overquota state and projects.

    Args:
        state: Overquota state.
        project_ids: List of uuids of projects.
        ex_project_ids: List of uuids of projects to be excluded.
        ex_user_ids: List of uuids of users to be excluded.

    Returns:
        Selected memberships.
    """
    q = Q()
    if ex_project_ids is not None:
        q &= ~Q(uuid__in=ex_project_ids)
    if project_ids:
        q &= Q(uuid__in=project_ids)
    projects = Project.objects.select_for_update().filter(q)
    q = Q()
    if ex_user_ids is not None:
        q &= ~Q(person__uuid__in=ex_user_ids)
    q &= Q(project__in=projects)
    return ProjectMembership.objects.select_related('project', 'person')\
        .filter(q & reduce(operator.or_,
                           [Q(overquota_state__startswith=state)
                            for state in states]))


def update_membership_state(updated_data):
    """
    This function updates the membership state for the given users and
    projects.

    Args:
        updated_data: A dictionary keyed by the overquota state to be updated
        that contains the information of project and user.
    """
    for state, pr_users in updated_data.iteritems():
        for project_id, user_ids in pr_users.iteritems():
            memberships = ProjectMembership.objects.filter(
                person__uuid__in=list(user_ids), project__uuid=project_id).\
                    exclude(overquota_state=state)
            now = datetime.now()
            if state == cf.OVERQUOTA_OK or state == cf.OVERQUOTA_OVERQUOTA:
                state_date = None if state == cf.OVERQUOTA_OK else now
                memberships.update(
                    overquota_state=state, overquota_state_date=now,
                    overquota_date=state_date)
            else:
                memberships.update(
                    overquota_state=state, overquota_state_date=now)


def update_overquota_state(user_ids, project_id, state, exclude=False):
    """
    This function updates the overquota state of project memberships
    accordingly.

    There are two possible scenarios:
    1) State should be updated to OVERQUOTA_OVERQUOTA to overquota users.
    2) State should be update to OVERQUOTA_OK to users who are not overquota.

    Args:
        user_ids: List of uuids of overquota users.
        project_id: UUID of project which is associated with overquota users.
        state: New state of membership.
        exclude: True if memberships of given users will not be updated;
        False otherwise.
    """
    now = datetime.now()
    OK = cf.OVERQUOTA_OK
    state_date = None if state == OK else now
    q = Q()
    q &= ~Q(person__uuid__in=user_ids) if exclude\
        else Q(person__uuid__in=user_ids)
    q &= Q(project__uuid=project_id)
    q &= ~Q(overquota_state=OK) if state == OK else Q(overquota_state=OK)
    ProjectMembership.objects.filter(q).update(
        overquota_state=state, overquota_state_date=now,
        overquota_date=state_date)


def get_reason(membership):
    """
    Get reason why a user associated with a specific membership is overquota.

    Typically, user is overquota, either due to deactication of a project they
    belong to (termination, suspension) or by being removed or suspened. If
    these are not the cases, then the project must have been modified and the
    limits of the resources have been decreased.

    Args:
        membership: Membership associated with the user who is overquota.
    Returns:
        True if the reason is associated with the status of the project; False
        otherwise.
        The actual reason.
    """
    project = membership.project
    if membership.state in ProjectMembership.DEACTIVATED_STATES:
        return (False,
                ProjectMembership.MEMBERSHIP_STATE_DISPLAY[membership.state])
    if project.state in Project.DEACTIVATED_STATES:
        return True, Project.O_STATE_DISPLAY[project.state]
    return True, 'Modified'


def _mk_quota_per_project(quotas, verbose=False, pairs=None):
    content_type = (lambda: defaultdict(dict)) if verbose else (lambda: [])
    overquotas_users = defaultdict(content_type)
    for (u_uuid, p_uuid, resource), usage in quotas.iteritems():
        project = p_uuid.split(qu.PROJECT_TAG)[1]
        user = u_uuid.split(qu.USER_TAG)[1]
        if pairs is not None and (user, project) not in pairs:
            continue
        if verbose:
            value = qu.from_holding(usage)
            overquotas_users[project][user][resource] = value
        else:
            overquotas_users[project].append(user)
    return overquotas_users


def get_overquota_users(memberships, services, resource=None, verbose=False):
    """
    Get a dictionary of overquota users per project.

    This function gets users who are using resources of specific services and
    surpass their current limit.

    Args:
        memberships: Memberships to check
        services: List of services whose resources are associated
        with, e.g. cyclades.
        resource: Resources to check if user is overquota.
        verbose: True if returned value should contain detailed info about
        resources usage.

    Returns:
        Dictionary keyed by project uuid. If verbose is `True`, it
        maps to detailed info on overquota users' resource limits and
        usage. If `False`, it maps to the list of overquota users.

    """
    users = set()
    projects = set()
    pairs = set()
    for membership in memberships:
        users.add(qu.get_user_ref(membership.person))
        projects.add(qu.get_project_ref(membership.project))
        pairs.add((membership.person.uuid, membership.project.uuid))
    flt = Q()
    flt &= Q(usage_max__gt=F("limit"))
    flt &= reduce(operator.or_, [Q(resource__startswith=service)
                                 for service in services])
    quotas = qu.qh.get_quota(holders=users, sources=projects,
                             resources=resource, flt=flt)
    return _mk_quota_per_project(
        quotas, verbose=verbose, pairs=pairs)


def is_membership_overquota(membership, services, resource=None):
    """
    This function checks if a specific membership is overquota.

    Args:
        membership: Membership to check if it's overquota.
        services: List of services which resources are associated
        with, e.g. cyclades.
        resource: Resources to check if membership is overquota.

    Returns:
        True if membership is overquota; False otherwise.
    """
    d = get_overquota_users([membership], services, resource)
    return bool(d)
