#!/usr/bin/env python
# Copyright (C) 2016-2017 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This script :
    - Reads from json files or gets the output of the commands:
        snf-manage stats-cyclades --cluster-details=False --output-format=json
        snf-manage stats-astakos --output-format=json
    - Parses the json and transforms it properly for graphite
    - Groups the data in packages of 100
    - Sends them to graphite with the use of pickle protocol
    - Prints these packages, if user wants to
"""

import sys
import socket
import subprocess
import pickle
import struct
import json
import collections
import datetime
import math
import argparse
import glob
import pprint
import textwrap


CONF = {}


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
        pass


def run_cmd():
    if CONF["type"] == "cyclades":
        cmd = "snf-manage stats-cyclades --cluster-details=False \
                --output-format=json".split()
    elif CONF["type"] == "astakos":
        cmd = "snf-manage stats-astakos --output-format=json".split()
    output = subprocess.check_output(cmd)
    return output


def set_conf(argv):
    """
    Reads command line arguments and sets some configuration/constants.
    No error handling.
    """
    # Help texts
    description = "Get synnefo stats and send them to a graphite installation."
    usage_examples = '''
    Usage examples
    --------------
    stats_to_graphite 127.0.0.1 --type=cyclades --prefix=stats --dc=DC1
    stats_to_graphite 127.0.0.1 --type=cyclades --input-files="../stats/*.json"
    stats_to_graphite 127.0.0.1 --port=2003 --type=astakos --prefix=stats
    '''
    server_help = "IP or hostname of carbon server."
    port_help = "Port Carbon's pickle receiver listens."
    type_help = 'Type can be "cyclades" or "astakos".'
    files_help = "Send data from json files. Each file must contain a json \
                  with the same format with the json output of \
                  snf-manage stats-* commands. If this has no value, the\
                  snf-manage command will be executed."
    prefix_help = "Prefix for the metrics. Prefixes are useful to group all \
                   metrics together."
    dc_help = "If dc is added the metrics will have the form: \
               prefix.dc.metric_name"
    verbose_help = "Print messages that are sent to graphite"

    parser = argparse.ArgumentParser(prog="stats_to_graphite",
                                     description=description,
                                     formatter_class=CustomFormatter,
                                     epilog=textwrap.dedent(usage_examples))

    parser.add_argument("server", help=server_help)
    parser.add_argument("--port", type=int, default=2004, help=port_help)
    parser.add_argument("--type", default="cyclades",
                        choices=["cyclades", "astakos"], help=type_help)
    parser.add_argument("--input-files", dest="input_files", metavar="PATTERN",
                        help=files_help)
    parser.add_argument("--prefix", default="okeanos", help=prefix_help)
    parser.add_argument("--dc", help=dc_help)
    parser.add_argument("--verbose", action="store_true", help=verbose_help)
    global CONF
    CONF = vars(parser.parse_args(argv))
    if (CONF["type"] != "cyclades") and (CONF["type"] != "astakos"):
        print 'Error: Type has to be "astakos" or "cyclades"'
        sys.exit(2)


def flatten(d, parent_key="", sep=".", replace_sep="_"):
    """
    Based on:
    https://stackoverflow.com/questions/6027558/flatten-nested-python-dictionaries-compressing-keys
    """
    items = []
    for k, v in d.items():
        # Only separator can be a dot. Replace all other dots.
        k = k.replace(sep, replace_sep)
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def remove_non_numerical_items(d):
    """
    Remove key-value pairs that the values cannot be represented in a timeline
    series
    """
    for k, v in d.items():
        if type(v).__name__ != "int":
            del d[k]
    return d


def bytes_readable(bytes):
    if bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB")
    i = int(math.floor(math.log(bytes, 1024)))
    p = math.pow(1024, i)
    s = int(round(bytes / p))
    return "%s%s" % (s, size_name[i])


def is_int(str):
    try:
        int(str)
        return True
    except ValueError:
        return False


def reword_keys(obj, prefix="okeanos"):
    sep = "."
    prefix = CONF["prefix"] or prefix
    if CONF["dc"] is not None:
        prefix = prefix + sep + CONF["dc"]
    prefix = prefix + sep + CONF["type"]
    if CONF["type"] == "cyclades":
        # prefix_total: Use for aggregated stats (not per cluster)
        prefix_total = "all"
        # No need for:
        #     - strings that include the service name
        #     - group all clusers together
        to_remove = ["_okeanos_grnet_gr", "clusters."]
        # disk and ram have in their keys their size in bytes
        transform_bytes = ["disk", "ram"]

        items = []
        for k, v in obj.items():
            if to_remove[0] in k:
                k = k.replace(to_remove[0], "")
            if transform_bytes[0] in k or transform_bytes[1] in k:
                last_dot_index = k.rfind(".")
                last_sub = k[last_dot_index+1:]

                if is_int(last_sub):
                    last_sub_transformed = bytes_readable(int(last_sub))
                    k = k.replace(last_sub, last_sub_transformed)
            # Images names: user:linux:debian -> user.linux.debian
            if k.startswith("images"):
                k = k.replace(":", sep)
            if not k.startswith("clusters"):
                k = prefix + sep + prefix_total + sep + k

            else:
                k = prefix + sep + k
            if to_remove[1] in k:
                k = k.replace(to_remove[1], "")
            items.append((k, v))
    elif CONF["type"] == "astakos":
        items = []
        for k, v in obj.items():
            k = prefix + sep + k
            items.append((k, v))

    return dict(items)


def parse_stats(flat_d):
    date = datetime.datetime.strptime(flat_d["datetime"], "%c")
    epoch_date = (date - datetime.datetime(1970, 1, 1)).total_seconds()
    parsed_stats = reword_keys(remove_non_numerical_items(flat_d))
    parsed_stats["datetime"] = epoch_date
    return parsed_stats


def send_messages(d, sock):
    parsed_stats = parse_stats(flatten(d))
    epoch_date = parsed_stats["datetime"]
    del parsed_stats["datetime"]
    if CONF["verbose"]:
        print "JSON items:", len(parsed_stats)
    tuples = ([])
    counter = 0
    logs = []
    json_items = len(parsed_stats)
    msgs = {}
    msgs["content"] = []
    msgs["logs"] = []
    for k, v in parsed_stats.items():
        tuples.append((k, (epoch_date, v)))
        if CONF["verbose"]:
            logs.append("%s %d %d" % (k, v, epoch_date))
        counter = counter + 1
        # Make packages of 100 metrics
        if counter % 100 == 0 or counter == json_items:
            msg = {}
            msg["package"] = pickle.dumps(tuples, 1)
            msg["size"] = struct.pack("!L", len(msg["package"]))
            tuples = ([])
            if CONF["verbose"]:
                pp = pprint.PrettyPrinter(indent=1)
                start_index = counter - len(logs) + 1
                print "Will send from %d to %d" % (start_index, counter)
                pp.pprint(logs)
                send_message(msg, sock)
                logs = []
                print "Sent"
                print "="*72
            else:
                send_message(msg, sock)
    return


def send_message(msg, sock):
    sock.sendall(msg["size"])
    sock.sendall(msg["package"])


def file_to_json(f):
    "Load json from file"
    with open(f) as data_file:
        try:
            data = json.load(data_file)
        except ValueError:
            return {}
    return data


def open_connection():
    "Open connection to graphite in order to send the messages"
    sock = socket.socket()
    try:
        sock.connect((CONF["server"], CONF["port"]))
    except socket.error:
        error_msg = textwrap.dedent("Couldn't connect to %(server)s on port %(port)d. Check carbon-cache.py.")
        raise SystemExit(error_msg % {"server": CONF["server"],
                                      "port": CONF["port"]})
    return sock


def main():
    argv = sys.argv[1:]
    set_conf(argv)
    logs = []
    if CONF["input_files"] is None:
        output = run_cmd()
        output = json.loads(output)
        sock = open_connection()
        send_messages(output, sock)
    else:
        file_names = glob.glob(CONF["input_files"])
        if(len(file_names) == 0):
            print "Wrong file pattern."
            sys.exit()
        if CONF["verbose"]:
            print "Will parse and send %d files" % len(file_names)
            print "="*72
        sock = open_connection()
        not_imported_files = []
        # For each file:
        #     - Read the included json
        #     - Transform its contents and send them to graphite
        for f in file_names:
            output = file_to_json(f)
            if "fault" in output or output.keys() == []:
                not_imported_files.append(f)
            else:
                send_messages(output, sock)
        pp = pprint.PrettyPrinter(indent=1)
        # If there files in not expected form, print their names
        if(len(not_imported_files) > 0):
            print "Error: files not sent:"
            pp.pprint(not_imported_files)
            print "Please, check that they have the expected json."


if __name__ == "__main__":
    main()
