# Copyright (C) 2016 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys
import logging
from astakos.im.models import AstakosUser, AstakosUserAuthProvider
from datetime import datetime, timedelta
from django.db.models import Q

from astakos.im import settings
from astakos.im.notifications import build_notification
from astakos.im import functions, quotas
from okeanos_policy.overquota import config as quota_cfg
from okeanos_policy.shibboleth import config as cfg

logger = logging.getLogger('shibboleth')


def unpack_state(state):
    category, sep, step = state.partition(cfg.SEP)
    step = int(step) if step else None
    return category, step


def pack_state(category, step):
    if category is None:
        return None
    s = category
    if step:
        s += "%s%s" % (cfg.SEP, step)
    return s


def get_next_state(category, step):
    policy = cfg.POLICY_CHAIN[category]['policy']
    next_cat = cfg.POLICY_CHAIN[category]['next']
    if policy is None or len(policy) <= step:
        return next_cat, 1
    return category, step+1


def get_due(category, step, epoch):
    if category is None:
        return sys.maxint
    policy = cfg.POLICY_CHAIN[category]['policy']
    if policy is None:
        return 0
    next_due = policy[step-1]
    now = datetime.now()
    elapsed = now - epoch
    delta = next_due - elapsed.days
    return delta


def str_to_date(s):
    return datetime.strptime(s, "%Y-%m-%d %H:%M:%S.%f")


def maybe_str_to_date(s):
    if not s:
        return s
    return str_to_date(s)


def get_shibboleth_auth(user):
    shib = user.get_auth_provider("shibboleth")
    return shib._instance


def get_shib_values(user):
    shib = get_shibboleth_auth(user)
    return get_shib_values_auth(shib)


def get_shib_values_auth(auth):
    info = auth.info
    return (
        info.get(cfg.SHIB_INFO_ACTION_NAME, ""),
        maybe_str_to_date(info.get(cfg.SHIB_INFO_DATE_NAME, "")),
        maybe_str_to_date(info.get(cfg.SHIB_INFO_EXPIRATION_DATE, ""))
    )


def decide(user, shib_values=None):
    if not shib_values:
        shib_values = get_shib_values(user)
    curr_state, curr_state_date, epoch = shib_values
    curr_category, curr_step = unpack_state(curr_state)
    next_category, next_step = get_next_state(curr_category, curr_step)
    due = get_due(next_category, next_step, epoch)
    return (curr_category, curr_step, next_category, next_step, epoch, due)


def send_notice(user, to_state, epoch):
    suspend_days = cfg.POLICY_CHAIN['SHIBBOLETH_SUSPEND']['policy'][0]
    shut_date = epoch + timedelta(days=suspend_days)
    sendemail(user.email, cfg.NOTIFICATION_SUBJ, cfg.WARN_TEMPLATE,
              {'user': user,
               'expire_in': cfg.SHIBBOLETH_EXPIRES_DAYS,
               'shut_date': shut_date.strftime("%d/%m/%Y"),
               'deletion_in': quota_cfg.HARD_ADMIN_DAYS[0]})
    logger.info("User %s notified: %s", user.email, to_state)


def set_local_account(user, active=False):
    if user.has_auth_provider('local'):
        l = user.get_auth_provider('local')
        l = l._instance
        l.active = active
        l.save()


def renew_token(user):
    user.renew_token()
    user.save()


SUSPENSION_REASON = 'shibboleth_expired'


def suspend(user, to_state, epoch):
    sendemail(user.email,
              cfg.LNOTICE_SUBJ_PREFIX + ' ' + cfg.NOTIFICATION_SUBJ,
              cfg.SUSPEND_TEMPLATE,
              {'user': user,
               'expire_in': cfg.SHIBBOLETH_EXPIRES_DAYS,
               'deletion_in': quota_cfg.HARD_ADMIN_DAYS[0]
               })
    functions.suspend_user_projects(user, reason=SUSPENSION_REASON)
    set_local_account(user, active=False)
    renew_token(user)
    logger.info("User %s suspended: %s", user.email, to_state)


def unsuspend(user, to_state, epoch):
    functions.unsuspend_user_projects(
        user, reason='shibboleth_recovered', condition=SUSPENSION_REASON)
    set_local_account(user, active=True)
    logger.info("User %s unsuspended", user.email)


def sendemail(recipient, subj, template, tvars):
    SENDER = settings.SERVER_EMAIL
    notification = build_notification(
        SENDER,
        [recipient],
        subj,
        template=template,
        dictionary=tvars)
    notification.send()


ACTION_DISPATCH = {
    'send_notice': send_notice,
    'suspend': suspend,
    'unsuspend': unsuspend,
}


def apply_actions(user, actions, state, epoch):
    for action in actions:
        func = ACTION_DISPATCH[action]
        func(user, state, epoch)


def check_user(user, dry_run, shib_values=None):
    curr_category, curr_step, \
        next_category, next_step, \
        epoch, due = decide(user, shib_values)
    curr_state = pack_state(curr_category, curr_step)
    next_state = pack_state(next_category, next_step)
    actions = cfg.POLICY_CHAIN.get(next_category, {}).get('actions', [])
    actions_str = ",".join(actions)
    if not(actions_str):
        actions_str = None
    due_str = ", due in: %s days" % due if next_state else ""
    logger.info("User %s current state: %s, next state: %s (action: %s)%s.",
                user.email, curr_state, next_state, actions_str, due_str)
    if due <= 0:
        if not dry_run:
            apply_actions(user, actions, next_state, epoch)
            now = datetime.now()
            updates = ((cfg.SHIB_INFO_ACTION_NAME, next_state),
                       (cfg.SHIB_INFO_DATE_NAME, str(now)))
            update_info(user, updates)


def recover_loggedin_user(user, dry_run):
    auth = get_shibboleth_auth(user)
    state, state_date, epoch = get_shib_values_auth(auth)
    category, step = unpack_state(state)
    actions = cfg.POLICY_CHAIN.get(category, {}).get('undo_actions', [])
    actions_str = ",".join(actions)
    logger.info("Recovering user %s (%s), last login: %s, current state: %s, "
                "action: %s",
                user.email, user.uuid, auth.last_login_at, state, actions_str)
    if not dry_run:
        for action in actions:
            func = ACTION_DISPATCH[action]
            func(user, None, epoch)
        keys = (cfg.SHIB_INFO_ACTION_NAME,
                cfg.SHIB_INFO_DATE_NAME,
                cfg.SHIB_INFO_EXPIRATION_DATE)
        del_info(user, keys)


def update_info(user, updates):
    shib = get_shibboleth_auth(user)
    return update_info_auth(shib, updates)


def update_info_auth(auth, updates):
    for key, val in updates:
        auth.info[key] = val
    auth.save()


def del_info(user, keys):
    auth = get_shibboleth_auth(user)
    for key in keys:
        auth.info.pop(key, None)
    auth.save()


MIGRATIONS = {
    cfg.FNOTICE: (cfg.SHIBBOLETH_NOTICE, 1, 0, False),
    cfg.SNOTICE: (cfg.SHIBBOLETH_NOTICE, 2, 5, False),
    cfg.LNOTICE: (cfg.SHIBBOLETH_NOTICE, 3, 8, False),
    cfg.SHUTDOWN: (cfg.SHIBBOLETH_SUSPEND, 1, 15, True),
    cfg.DELETE: (cfg.SHIBBOLETH_DELETED, None, 22, False),
}


def migrate_user(user, dry_run):
    auth = get_shibboleth_auth(user)
    curr_state, curr_state_date, epoch = get_shib_values_auth(auth)
    try:
        new_category, new_step, elapsed, execute = MIGRATIONS[curr_state]
    except KeyError:
        logger.info("Can't migrate %s" % auth)

    if not epoch:
        epoch = curr_state_date - timedelta(days=elapsed)
    new_state = pack_state(new_category, new_step)
    updates = ((cfg.SHIB_INFO_ACTION_NAME, new_state),
               (cfg.SHIB_INFO_EXPIRATION_DATE, str(epoch)))
    logger.info("Migrating user %s: from state: %s to state: %s",
                user.email, curr_state, new_state)
    if not dry_run:
        if execute:
            actions = cfg.POLICY_CHAIN.get(new_category, {}).get('actions', [])
            apply_actions(user, actions, new_state, epoch)
        update_info_auth(auth, updates)


def initialize_expired_auth(auth, dry_run):
    curr_state, curr_state_date, epoch = get_shib_values_auth(auth)
    if not curr_state:
        logger.info("Setting user %s: %s",
                    auth.user.email, cfg.SHIBBOLETH_EXPIRED)
        now = datetime.now()
        keys = (cfg.SHIB_INFO_ACTION_NAME,
                cfg.SHIB_INFO_DATE_NAME,
                cfg.SHIB_INFO_EXPIRATION_DATE)
        values = (cfg.SHIBBOLETH_EXPIRED, now, now)
        updates = zip(keys, map(str, values))
        if dry_run:
            return values
        update_info_auth(auth, updates)
    return None


def mk_old_states_query():
    q = Q()
    for state in MIGRATIONS.keys():
        q |= Q(info_data__contains=state)
    return q


def get_users_to_migrate(exclude=None):
    objects = AstakosUserAuthProvider.objects.select_for_update()
    old_states_q = mk_old_states_query()
    userids = objects.filter(module='shibboleth').filter(
        old_states_q).values_list('user_id', flat=True)

    objects = AstakosUser.objects.select_for_update().select_related()
    if exclude:
        objects = objects.exclude(pk__in=exclude)
    users = objects.filter(pk__in=userids)
    logger.info("Found %s users in deprecated states", len(users))
    return users


def get_loggedin_users():
    days = cfg.SHIBBOLETH_EXPIRES_DAYS
    login_before = datetime.now() - timedelta(days=days)
    objects = AstakosUserAuthProvider.objects.select_for_update()
    userids = objects.filter(module='shibboleth').filter(
        last_login_at__gt=login_before,
        info_data__contains=cfg.SHIB_INFO_ACTION_NAME
        ).values_list('user_id', flat=True)

    objects = AstakosUser.objects.select_for_update().select_related()
    objects = objects.exclude(groups__name=cfg.SHIBBOLETH_EXCEPTIONS_GRP)
    users = objects.filter(is_active=True, pk__in=userids)
    logger.info("Found %s users that have re-logged in using shibboleth "
                "in the last %s days.", len(users), days)
    return users


def get_expired_users():
    days = cfg.SHIBBOLETH_EXPIRES_DAYS
    login_before = datetime.now() - timedelta(days=days)
    objects = AstakosUserAuthProvider.objects.select_for_update()
    userids = objects.filter(module='shibboleth').filter(
        Q(last_login_at__lt=login_before) |
        Q(last_login_at__isnull=True, created__lt=login_before)
        ).values_list('user_id', flat=True)

    objects = AstakosUser.objects.select_for_update().select_related()
    objects = objects.exclude(groups__name=cfg.SHIBBOLETH_EXCEPTIONS_GRP)
    users = objects.filter(is_active=True, pk__in=userids)

    user_uuids_with_quotas = set()
    user_quotas = quotas.get_users_quotas(
        users, cfg.RESOURCES, flt=Q(usage_max__gt=0))
    for user, quota in user_quotas.iteritems():
        if quota:
            user_uuids_with_quotas.add(user)

    logger.info("Found %s shibboleth-expired users that use resources",
                len(user_uuids_with_quotas))
    return [user for user in users
            if user.uuid in user_uuids_with_quotas]
