# Copyright (C) 2016 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
os.environ['DJANGO_SETTINGS_MODULE'] = 'synnefo.settings'
import mock
from collections import OrderedDict
from datetime import datetime, timedelta
import unittest
from  okeanos_policy.overquota import config
from okeanos_policy.overquota.quota_policy import QuotaPolicy


config.SOFT_USER_DAYS = [0, 2]
config.SOFT_ADMIN_DAYS = [3]
config.HARD_USER_DAYS = [4]
config.HARD_ADMIN_DAYS = [5]

config.POLICY_CHAIN['SOFT_USER']['policy'] = config.SOFT_USER_DAYS
config.POLICY_CHAIN['SOFT_ADMIN']['policy'] = config.SOFT_ADMIN_DAYS
config.POLICY_CHAIN['HARD_USER']['policy'] = config.HARD_USER_DAYS
config.POLICY_CHAIN['HARD_ADMIN']['policy'] = config.HARD_ADMIN_DAYS


STATES = config.POLICY_CHAIN.keys()


def init_mock_quota_policy(**kwargs):
    return mock.MagicMock(
        current_category=kwargs.get('current_category', None),
        transitions=kwargs.get('transitions', None),
        last_update=kwargs.get('last_update', None),
        repeat=kwargs.get('repeat', None),
        overquota_resources=kwargs.get('overquota_resources', None),
        next_category=kwargs.get('next_category', None),
        get_next_state_category=QuotaPolicy.__dict__['get_next_state_category'],
        set_num_of_transitions=QuotaPolicy.__dict__['set_num_of_transitions'],
        get_time_next_state=QuotaPolicy.__dict__['get_time_next_state'],
        get_time_period=QuotaPolicy.__dict__['get_time_period'],
        define_resource_actions=QuotaPolicy.__dict__['define_resource_actions']
    )


class TestQuotaPolicy(unittest.TestCase):
    def test_get_next_state_category(self):
        mock_qp = init_mock_quota_policy(
            current_category=config.OVERQUOTA_OK, transitions=0)
        self.assertEqual(mock_qp.get_next_state_category(mock_qp),
                         STATES[STATES.index(config.OVERQUOTA_OK) + 1])
        state = config.SEVERITY[0] + '_' + config.RECIPIENT[0]
        mock_qp = init_mock_quota_policy(
            current_category=state, transitions=1)
        self.assertEqual(mock_qp.get_next_state_category(mock_qp),
                         state)
        mock_qp = init_mock_quota_policy(
            current_category=state, transitions=2)
        self.assertEqual(mock_qp.get_next_state_category(mock_qp),
                         STATES[STATES.index(state) + 1])
        mock_qp = init_mock_quota_policy(
            current_category=state, transitions=2, repeat=True)
        self.assertEqual(mock_qp.get_next_state_category(mock_qp),
                         state)

    def test_set_num_of_transitions(self):
        mock_qp = init_mock_quota_policy(
            current_category=config.OVERQUOTA_OK, transitions=0, repeat=True,
            next_category=config.OVERQUOTA_OK)
        mock_qp.set_num_of_transitions(mock_qp)
        self.assertEqual(mock_qp.transitions, 0)
        mock_qp = init_mock_quota_policy(
            current_category=config.OVERQUOTA_OK, transitions=0, repeat=False,
            next_category=STATES[STATES.index(config.OVERQUOTA_OK) + 1])
        mock_qp.set_num_of_transitions(mock_qp)
        self.assertEqual(mock_qp.transitions, None)
        state = config.SEVERITY[0] + '_' + config.RECIPIENT[0]
        mock_qp = init_mock_quota_policy(
            current_category=state, transitions=1,
            next_category=state)
        mock_qp.set_num_of_transitions(mock_qp)
        self.assertEqual(mock_qp.transitions, 2)
        mock_qp = init_mock_quota_policy(
            current_category=state, transitions=3,
            next_category=state)
        mock_qp.set_num_of_transitions(mock_qp)
        self.assertEqual(mock_qp.transitions, 2)
        mock_qp = init_mock_quota_policy(
            current_category=state, transitions=3,
            next_category=STATES[STATES.index(state) + 1])
        mock_qp.set_num_of_transitions(mock_qp)
        self.assertEqual(mock_qp.transitions, 1)

    def test_get_time_next_state(self):
        now = datetime.now()
        mock_qp = init_mock_quota_policy(last_update=now,
                                         next_category=config.OVERQUOTA_OK)
        self.assertEqual(mock_qp.get_time_next_state(mock_qp), 0)
        state = config.SEVERITY[0] + '_' + config.RECIPIENT[0]
        mock_qp = init_mock_quota_policy(
            last_update=now, next_category=state, transitions=2)
        self.assertEqual(mock_qp.get_time_next_state(mock_qp), 2)
        date = now + timedelta(days=-3)
        mock_qp = init_mock_quota_policy(
            last_update=date, next_category=state, transitions=2)
        self.assertEqual(mock_qp.get_time_next_state(mock_qp), 0)

    def test_get_time_period(self):
        now = datetime.now()
        mock_qp = init_mock_quota_policy(
            last_update=now, next_category=config.OVERQUOTA_OVERQUOTA,
            overquota_resources=['cyclades.ram'])
        mock_qp.get_time_period(mock_qp)
        self.assertEqual(mock_qp._calculate_time_period.call_count, 1)
        mock_qp = init_mock_quota_policy(
            last_update=now, next_category=config.OVERQUOTA_OVERQUOTA,
            overquota_resources=['cyclades.vm'])
        mock_qp.get_time_period(mock_qp)
        self.assertEqual(mock_qp._calculate_time_period.call_count, 2)

    def test_define_resource_actions(self):
        state = config.SEVERITY[0] + '_' + config.RECIPIENT[0]
        mock_qp = init_mock_quota_policy(
            next_category=config.OVERQUOTA_OVERQUOTA,
            overquota_resources=['cyclades.ram'])
        soft_actions, hard_actions = mock_qp.define_resource_actions(mock_qp)
        self.assertEqual(soft_actions, set())
        self.assertEqual(hard_actions, set())
        mock_qp = init_mock_quota_policy(
            next_category=state,
            overquota_resources=['cyclades.ram'])
        soft_actions, hard_actions = mock_qp.define_resource_actions(mock_qp)
        self.assertEqual(soft_actions,
                         {config.RESOURCE_ACTIONS[config.SEVERITY[0],
                         'cyclades.ram']})
        self.assertEqual(hard_actions, set())


if __name__ == '__main__':
    unittest.main()
