# Copyright (C) 2016 GRNET S.A.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.test import TestCase
from astakos.im.tests.common import get_local_user
from okeanos_policy.shibboleth import enforce
from datetime import datetime, timedelta
import mock


class TestShibbolethEnforce(TestCase):
    def setUp(self):
        self.user1 = get_local_user("user1@synnefo.org")
        self.user1.add_auth_provider(module='shibboleth')
        self.user2 = get_local_user("user2@synnefo.org")
        self.user2.add_auth_provider(module='shibboleth')
        self.user3 = get_local_user("user3@synnefo.org")
        self.users = [self.user1, self.user2, self.user3]
        self.uuids = [user.uuid for user in self.users]

    def nfu(self, users):
        return sorted([u.email for u in users])

    def test_expired_single_per_user(self):
        enforce.cfg.SHIBBOLETH_EXPIRES_DAYS = 0
        with mock.patch("astakos.im.quotas.get_users_quotas") as mk:
            mk.return_value = dict(zip(self.uuids, [True] * (len(self.uuids))))
            expired = enforce.get_expired_users()
            self.assertEqual(self.nfu(expired),
                             self.nfu([self.user1, self.user2]))

    def test_expired_check_shibboleth_only(self):
        enforce.cfg.SHIBBOLETH_EXPIRES_DAYS = 1

        # user1 local provider is old
        inst = self.user1.get_auth_provider('local')._instance
        inst.created = datetime.now() - timedelta(days=2)
        inst.save()

        # user2 shibboleth provider is old
        inst = self.user2.get_auth_provider('shibboleth')._instance
        inst.created = datetime.now() - timedelta(days=2)
        inst.save()
        with mock.patch("astakos.im.quotas.get_users_quotas") as mk:
            mk.return_value = dict(zip(self.uuids, [True] * (len(self.uuids))))
            expired = enforce.get_expired_users()
            self.assertEqual(self.nfu(expired), self.nfu([self.user2]))

    def test_loggedin_single_per_user(self):
        enforce.cfg.SHIBBOLETH_EXPIRES_DAYS = 60
        # user1 has re-logged in using both local and shibboleth
        inst = self.user1.get_auth_provider('local')._instance
        inst.last_login_at = datetime.now()
        inst.info[enforce.cfg.SHIB_INFO_ACTION_NAME] = 'SHIBBOLETH_EXPIRED'
        inst.save()

        inst = self.user1.get_auth_provider('shibboleth')._instance
        inst.last_login_at = datetime.now()
        inst.info[enforce.cfg.SHIB_INFO_ACTION_NAME] = 'SHIBBOLETH_EXPIRED'
        inst.save()
        reloggedin = enforce.get_loggedin_users()
        self.assertEqual(self.nfu(reloggedin), self.nfu([self.user1]))

    def test_loggedin_check_shibboleth_only(self):
        enforce.cfg.SHIBBOLETH_EXPIRES_DAYS = 60
        # user1 has re-logged in using local shibboleth
        inst = self.user1.get_auth_provider('local')._instance
        inst.last_login_at = datetime.now()
        inst.info[enforce.cfg.SHIB_INFO_ACTION_NAME] = 'SHIBBOLETH_EXPIRED'
        inst.save()

        # user2 has re-logged in using shibboleth
        inst = self.user2.get_auth_provider('shibboleth')._instance
        inst.last_login_at = datetime.now()
        inst.info[enforce.cfg.SHIB_INFO_ACTION_NAME] = 'SHIBBOLETH_EXPIRED'
        inst.save()
        reloggedin = enforce.get_loggedin_users()
        self.assertEqual(self.nfu(reloggedin), self.nfu([self.user2]))

    def test_get_migration_single_per_user(self):
        inst = self.user1.get_auth_provider('local')._instance
        inst.info[enforce.cfg.SHIB_INFO_ACTION_NAME] = 'fnotice'
        inst.save()

        inst = self.user1.get_auth_provider('shibboleth')._instance
        inst.info[enforce.cfg.SHIB_INFO_ACTION_NAME] = 'fnotice'
        inst.save()
        migr = enforce.get_users_to_migrate()
        self.assertEqual(self.nfu(migr), self.nfu([self.user1]))

    def test_get_migration_shibboleth_only(self):
        # user1 local provider has deprecated data
        inst = self.user1.get_auth_provider('local')._instance
        inst.info[enforce.cfg.SHIB_INFO_ACTION_NAME] = 'fnotice'
        inst.save()

        # user2 shibboleth provider has deprecated data
        inst = self.user2.get_auth_provider('shibboleth')._instance
        inst.info[enforce.cfg.SHIB_INFO_ACTION_NAME] = 'fnotice'
        inst.save()

        migr = enforce.get_users_to_migrate()
        self.assertEqual(self.nfu(migr), self.nfu([self.user2]))
